import { createElement, TagsWithInnerContent } from 'zhead';

function setAttributesWithSideEffects(head, $el, entry, tag) {
  const attrs = tag.props || {};
  const sdeKey = `${tag._p}:attr`;
  Object.entries(entry._sde).filter(([key]) => key.startsWith(sdeKey)).forEach(([key, fn]) => {
    delete entry._sde[key] && fn();
  });
  Object.entries(attrs).forEach(([k, value]) => {
    value = String(value);
    const attrSdeKey = `${sdeKey}:${k}`;
    head._removeQueuedSideEffect(attrSdeKey);
    if (k === "class") {
      for (const c of value.split(" ")) {
        if (!$el.classList.contains(c)) {
          $el.classList.add(c);
          head._removeQueuedSideEffect(`${attrSdeKey}:${c}`);
          entry._sde[`${attrSdeKey}:${c}`] = () => $el.classList.remove(c);
        }
      }
      return;
    }
    if ($el.getAttribute(k) !== value) {
      $el.setAttribute(k, value);
      if (!k.startsWith("data-h-"))
        entry._sde[attrSdeKey] = () => $el.removeAttribute(k);
    }
  });
}

async function renderDOMHead(head, options = {}) {
  const dom = options.document || window.document;
  const tags = await head.resolveTags();
  await head.hooks.callHook("dom:beforeRender", { head, tags, document: dom });
  for (const tag of tags) {
    const entry = head.headEntries().find((e) => e._i === Number(tag._e));
    const sdeKey = `${tag._s || tag._p}:el`;
    const $newEl = createElement(tag, dom);
    const $el = tag._s ? dom.querySelector(`[${tag._s}]`) : null;
    const renderCtx = { tag, document: dom, head };
    await head.hooks.callHook("dom:renderTag", renderCtx);
    if ($el) {
      head._removeQueuedSideEffect(sdeKey);
      if ($newEl.isEqualNode($el))
        continue;
      if (Object.keys(tag.props).length === 0) {
        $el.remove();
        continue;
      }
      setAttributesWithSideEffects(head, $el, entry, tag);
      if (TagsWithInnerContent.includes(tag.tag))
        $el.innerHTML = tag.children || "";
      entry._sde[sdeKey] = () => $el?.remove();
      continue;
    }
    if (tag.tag === "title" && tag.children) {
      dom.title = tag.children;
      continue;
    }
    if (tag.tag === "htmlAttrs" || tag.tag === "bodyAttrs") {
      setAttributesWithSideEffects(head, dom[tag.tag === "htmlAttrs" ? "documentElement" : "body"], entry, tag);
      continue;
    }
    switch (tag.tagPosition) {
      case "bodyClose":
        dom.body.appendChild($newEl);
        break;
      case "bodyOpen":
        dom.body.insertBefore($newEl, dom.body.firstChild);
        break;
      case "head":
      default:
        dom.head.appendChild($newEl);
        break;
    }
    entry._sde[sdeKey] = () => $newEl?.remove();
  }
  head._flushQueuedSideEffects();
}
let domUpdatePromise = null;
async function debouncedRenderDOMHead(delayedFn, head, options = {}) {
  function doDomUpdate() {
    domUpdatePromise = null;
    return renderDOMHead(head, options);
  }
  return domUpdatePromise = domUpdatePromise || new Promise((resolve) => delayedFn(() => resolve(doDomUpdate())));
}

export { debouncedRenderDOMHead, domUpdatePromise, renderDOMHead };
